-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved
local g_modName = g_currentModName
local g_modPath = g_currentModDirectory
if FillTypeManager.SEND_NUM_BITS < 10 then
    FillTypeManager.SEND_NUM_BITS = 10
    printf("%s: FillType limit increased to 1023", g_modName)
end
local THLimitBreak = {}
THLimitBreak.extraGroundTextures = {
    byId = {},
    byIndex = {},
    byValue = {}
}
THLimitBreak.numVanillaSprayTextures = 0
for _, v in pairs(FieldSprayType.getAll()) do
    if v > THLimitBreak.numVanillaSprayTextures then
        THLimitBreak.numVanillaSprayTextures = v
    end
end
THLimitBreak.extraSprayTextures = {
    byId = {},
    byIndex = {},
    byValue = {}
}
THLimitBreak.numVanillaChopperTextures = 0
for _, v in pairs(FieldChopperType) do
    if v > THLimitBreak.numVanillaChopperTextures then
        THLimitBreak.numVanillaChopperTextures = v
    end
end
THLimitBreak.extraChopperTextures = {
    byId = {},
    byIndex = {},
    byValue = {}
}
function THLimitBreak.displayMsg(text, ...)
    text = string.format(text, ...)
    printf("%s: %s", g_modName, text)
end
function THLimitBreak.errorMsg(text, ...)
    THLimitBreak.displayMsg("ERROR: " .. text, ...)
    printCallstack()
end
function THLimitBreak.getNextGroundTextureIndex()
    local nextIndex = 0
    for _, v in pairs(FieldGroundType.getAll()) do
        nextIndex = math.max(v, nextIndex)
    end
    nextIndex = nextIndex + 1
    return nextIndex
end
function THLimitBreak.getNextSprayTextureIndex()
    local nextIndex = 0
    for _, v in pairs(FieldSprayType.getAll()) do
        nextIndex = math.max(v, nextIndex)
    end
    nextIndex = nextIndex + 1
    return nextIndex
end
function THLimitBreak.getNextChopperTextureIndex()
    local nextIndex = 0
    for _, v in pairs(FieldChopperType) do
        nextIndex = math.max(v, nextIndex)
    end
    nextIndex = nextIndex + 1
    return nextIndex
end
function THLimitBreak.hook_getFieldGroundTypeName(value, superFunc, ...)
    local function appendFunc(rName, ...)
        if rName == nil and value ~= nil then
            local groundTextureData = THLimitBreak.extraGroundTextures.byValue[value]
            if groundTextureData ~= nil then
                rName = groundTextureData.id
            end
        end
        return rName, ...
    end
    return appendFunc(superFunc(value, ...))
end
FieldGroundType.getName = Utils.overwrittenFunction(FieldGroundType.getName, THLimitBreak.hook_getFieldGroundTypeName)
function THLimitBreak.hook_getFieldSprayTypeName(value, superFunc, ...)
    local function appendFunc(rName, ...)
        if rName == nil and value ~= nil then
            local sprayTextureData = THLimitBreak.extraSprayTextures.byValue[value]
            if sprayTextureData ~= nil then
                rName = sprayTextureData.id
            end
        end
        return rName, ...
    end
    return appendFunc(superFunc(value, ...))
end
FieldSprayType.getName = Utils.overwrittenFunction(FieldSprayType.getName, THLimitBreak.hook_getFieldSprayTypeName)
function THLimitBreak.hook_loadFieldGroundTypes(manager, superFunc, filename, ...)
    local xmlBaseKey = "fieldGround"
    local densityMapsKey = string.format("%s.densityMaps", xmlBaseKey)
    local limitBreakKey = string.format("%s.thLimitBreak", densityMapsKey)
    local xmlFile = nil
    if filename ~= nil and filename ~= "" then
        xmlFile = XMLFile.load("THFieldGroundXML", filename)
        if xmlFile ~= nil then
            local extraGroundTypesKey = string.format("%s.extraGroundTypes", limitBreakKey)
            local allGroundTypes = FieldGroundType.getAll()
            local orderedGroundTypes = FieldGroundType.getAllOrdered()
            if xmlFile:hasProperty(extraGroundTypesKey) then
                xmlFile:iterate(extraGroundTypesKey .. ".groundType", function(_, pGroundTypeKey)
                    local groundTypeName = xmlFile:getString(pGroundTypeKey .. "#name", "")
                    local groundTypeId = groundTypeName:upper()
                    local groundTypeValue = THLimitBreak.getNextGroundTextureIndex()
                    if groundTypeName == "" then
                        THLimitBreak.errorMsg("Invalid groundType name: %s", groundTypeName)
                    elseif groundTypeValue == nil or groundTypeValue <= 1 then
                        THLimitBreak.errorMsg("Invalid groundType value: %s", groundTypeValue)
                    elseif FieldGroundType.getByName(groundTypeName) ~= nil
                        or FieldGroundType.getName(groundTypeValue) ~= nil
                        or allGroundTypes[groundTypeId] ~= nil
                    then
                        THLimitBreak.errorMsg("GroundType %q already exists", groundTypeName)
                    else
                        FieldGroundType[groundTypeId] = groundTypeValue
                        allGroundTypes[groundTypeId] = groundTypeValue
                        table.insert(orderedGroundTypes, groundTypeValue)
                        local groundTextureData = {
                            id = groundTypeId,
                            name = groundTypeName,
                            index = #THLimitBreak.extraGroundTextures.byIndex + 1,
                            value = groundTypeValue
                        }
                        THLimitBreak.extraGroundTextures.byId[groundTextureData.id] = groundTextureData
                        THLimitBreak.extraGroundTextures.byIndex[groundTextureData.index] = groundTextureData
                        THLimitBreak.extraGroundTextures.byValue[groundTextureData.value] = groundTextureData
                    end
                end)
                table.sort(orderedGroundTypes)
            end
            local extraSprayTypesKey = string.format("%s.extraSprayTypes", limitBreakKey)
            local allSprayTypes = FieldSprayType.getAll()
            local orderedSprayTypes = FieldSprayType.getAllOrdered()
            if xmlFile:hasProperty(extraSprayTypesKey) then
                xmlFile:iterate(extraSprayTypesKey .. ".sprayType", function(_, pSprayTypeKey)
                    local sprayTypeName = xmlFile:getString(pSprayTypeKey .. "#name", "")
                    local sprayTypeId = sprayTypeName:upper()
                    local sprayTypeValue = THLimitBreak.getNextSprayTextureIndex()
                    if sprayTypeName == "" then
                        THLimitBreak.errorMsg("Invalid sprayType name: %s", sprayTypeName)
                    elseif sprayTypeValue == nil or sprayTypeValue <= 1 then
                        THLimitBreak.errorMsg("Invalid sprayType value: %s", sprayTypeValue)
                    elseif FieldSprayType.getByName(sprayTypeName) ~= nil
                        or FieldSprayType.getName(sprayTypeValue) ~= nil
                        or allSprayTypes[sprayTypeId] ~= nil
                    then
                        THLimitBreak.errorMsg("SprayType %q already exists", sprayTypeName)
                    else
                        FieldSprayType[sprayTypeId] = sprayTypeValue
                        allSprayTypes[sprayTypeId] = sprayTypeValue
                        table.insert(orderedSprayTypes, sprayTypeValue)
                        local sprayTextureData = {
                            id = sprayTypeId,
                            name = sprayTypeName,
                            index = #THLimitBreak.extraSprayTextures.byIndex + 1,
                            value = sprayTypeValue
                        }
                        THLimitBreak.extraSprayTextures.byId[sprayTextureData.id] = sprayTextureData
                        THLimitBreak.extraSprayTextures.byIndex[sprayTextureData.index] = sprayTextureData
                        THLimitBreak.extraSprayTextures.byValue[sprayTextureData.value] = sprayTextureData
                    end
                end)
                table.sort(orderedSprayTypes)
            end
            local extraChopperTypesKey = string.format("%s.extraChopperTypes", limitBreakKey)
            if xmlFile:hasProperty(extraChopperTypesKey) then
                xmlFile:iterate(extraChopperTypesKey .. ".chopperType", function(_, pChopperTypeKey)
                    local chopperTypeName = xmlFile:getString(pChopperTypeKey .. "#name", "")
                    local chopperTypeId = chopperTypeName:upper()
                    local chopperTypeValue = THLimitBreak.getNextChopperTextureIndex()
                    if chopperTypeName == "" then
                        THLimitBreak.errorMsg("Invalid chopperType name: %s", chopperTypeName)
                    elseif chopperTypeValue == nil or chopperTypeValue <= 1 then
                        THLimitBreak.errorMsg("Invalid chopperType value: %s", chopperTypeValue)
                    elseif FieldChopperType[chopperTypeId] ~= nil then
                        THLimitBreak.errorMsg("ChopperType %q already exists", chopperTypeName)
                    else
                        FieldChopperType[chopperTypeId] = chopperTypeValue
                        local chopperTextureData = {
                            id = chopperTypeId,
                            name = chopperTypeName,
                            index = #THLimitBreak.extraChopperTextures.byIndex + 1,
                            value = chopperTypeValue
                        }
                        THLimitBreak.extraChopperTextures.byId[chopperTextureData.id] = chopperTextureData
                        THLimitBreak.extraChopperTextures.byIndex[chopperTextureData.index] = chopperTextureData
                        THLimitBreak.extraChopperTextures.byValue[chopperTextureData.value] = chopperTextureData
                    end
                end)
            end
        end
    end
    local function appendFunc(...)
        if xmlFile ~= nil then
            for _, groundTextureData in ipairs(THLimitBreak.extraGroundTextures.byIndex) do
                local groundTypeXMLKey = string.format("%s.groundTypes.%s", densityMapsKey, groundTextureData.name)
                manager:loadGroundIdFromXML(groundTextureData.value, xmlFile, groundTypeXMLKey, groundTextureData.value, { 1, 1, 1, 1 })
            end
            for _, sprayTextureData in ipairs(THLimitBreak.extraSprayTextures.byIndex) do
                local sprayTypeXMLKey = string.format("%s.sprayTypes.%s", densityMapsKey, sprayTextureData.name)
                manager:loadSprayIdFromXML(sprayTextureData.value, xmlFile, sprayTypeXMLKey, sprayTextureData.value, { 1, 1, 1, 1 })
            end
            for _, chopperTextureData in ipairs(THLimitBreak.extraChopperTextures.byIndex) do
                local chopperTypeXMLKey = string.format("%s.sprayTypes.%s", densityMapsKey, chopperTextureData.name)
                local defaultChopperValue = chopperTextureData.value + THLimitBreak.numVanillaSprayTextures + THLimitBreak.numVanillaChopperTextures
                manager:loadChopperIdFromXML(chopperTextureData.value, xmlFile, chopperTypeXMLKey, defaultChopperValue, { 1, 1, 1, 1 })
            end
            xmlFile:delete()
        end
    end
    return appendFunc(superFunc(manager, filename, ...))
end
FieldGroundSystem.loadGroundTypes = Utils.overwrittenFunction(FieldGroundSystem.loadGroundTypes, THLimitBreak.hook_loadFieldGroundTypes)